package JointFeldmanProtocol;

import FeldmanVerifiableSecretSharing.ShamirSecretSharing.Polynomial;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import org.bouncycastle.util.Arrays;
import org.factcenter.qilin.primitives.concrete.Zpstar;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 2/5/2016.
 *
 * an implementation of a version of Pedersen's distributed key generation protocol
 */
public class DKG extends VerifiableSecretSharing implements Runnable{

    private Network.User user;
    private Handler handler;
    private Polynomial.Point[] shares;
    private BigInteger[][] commitmentsArray;
    private BigInteger[] ys;
    private Set<Integer> QUAL;

    private BigInteger x;
    private BigInteger y;
    private BigInteger[] commitments;

    public DKG(int t, int n, BigInteger x, Random random, BigInteger p, BigInteger q, BigInteger g,Network network) {
        super(t, n, x, random, p, q, g);
        this.commitmentsArray = new BigInteger[n][t + 1];
        this.shares = new Polynomial.Point[n];
        this.handler = new Handler();
        this.user = network.connect(handler);
    }


    private void stage1(){
        int n = getN();
        int t = getT();
        int i = user.getID();
        BigInteger[] commitments = super.getCommitments();
        System.arraycopy(commitments, 0, commitmentsArray[i - 1], 0, commitmentsArray[i - 1].length);

        Network.CommitmentMessage commitment;
        for (int j = 1; j <= commitmentsArray[i - 1].length; j ++){
            commitment = new Network.CommitmentMessage(j,commitmentsArray[i - 1][j - 1]);
            user.broadcast(commitment);
        }

        Network.SecretMessage secret;
        for (int j = 1; j <= n ; j++ ){
            if(j != i){
                secret = new Network.SecretMessage(getShare(j));
                user.send(j,secret);
            }
        }
        while (handler.secretsCounter < n - 1 || handler.commitmentsCounter < (n - 1) * (t + 1)){
            try {
                Thread.sleep(30);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
    }

    private void stage2(){
        int n = getN();
        BigInteger g = getGenerator();
        Zpstar zpstar = getZpstar();
        int i = user.getID();
        QUAL = new HashSet<Integer>();
        Network.Message message = null;
        for (int j = 1; j <= n ; j++ ){
            if(zpstar.multiply(g,shares[j - 1].y).equals(verify(j,commitmentsArray[j],zpstar))){
                QUAL.add(j);
            }else{
                //message = new Message(Type.Complaint, j)
                user.broadcast(message);
            }
        }
        //sending y after Complaints
        Network.YMessage yMessage = new Network.YMessage(super.getY());
        user.broadcast(yMessage);
    }

    private void stage3(){
        int n = getN();
        while (handler.ysCounter < n - 1 ){
            try {
                Thread.sleep(30);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
        //Todo receive something private from each complaint + send what necessary
    }

    private void stage4(){
        int t = getT();
        Zpstar zpstar = getZpstar();
        BigInteger y = zpstar.zero();
        for (int i : QUAL) {
            y = zpstar.add(y , ys[i - 1]);
        }
        this.commitments = new BigInteger[t];
        BigInteger commitment;
        for (int k = 1; k <= t ; k++){
            commitment = zpstar.zero();
            for (int i : QUAL) {
                commitment = zpstar.add(commitment,commitmentsArray[i - 1][k - 1]);
            }
            commitments[k - 1] = commitment;
        }

        int j = user.getID();
        BigInteger x = BigInteger.ZERO;
        for (int i : QUAL) {
            x = x.add(shares[i - 1].y);
        }
        this.x = x.mod(getQ());
    }

    @Override
    public BigInteger getY() {
        return y;
    }

    @Override
    public BigInteger[] getCommitments() {
        return Arrays.clone(commitments);
    }

    @Override
    protected Polynomial.Point getShare(int i) {
        return null;
    }

    @Override
    public void run() {
        stage1();
        stage2();
        stage3();
        stage4();
    }

    private class Handler implements Network.MailHandler {

        final int id;
        int secretsCounter;
        int commitmentsCounter;
        int ysCounter;

        private Handler() {
            this.id = user.getID();
            this.secretsCounter = 0;
            this.commitmentsCounter = 0;
            this.ysCounter = 0;
        }

        @Override
        public void handel(Network.Mail mail) {
            if(mail.isPrivate){
                if(mail.message instanceof Network.SecretMessage){
                    Polynomial.Point secret = ((Network.SecretMessage)mail.message).secret;
                    if(shares[id - 1] == null) {
                        shares[id - 1] = secret;
                        secretsCounter++;
                    }
                }
            }else{
                if(mail.message instanceof Network.CommitmentMessage){
                    Network.CommitmentMessage commitmentMessage = (Network.CommitmentMessage)mail.message;
                    if(commitmentsArray[mail.senderID - 1][commitmentMessage.k] == null) {
                        commitmentsArray[mail.senderID - 1][commitmentMessage.k] = commitmentMessage.commitment;
                        commitmentsCounter++;
                    }
                }else if(mail.message instanceof Network.YMessage){
                    BigInteger y = ((Network.YMessage)mail.message).y;
                    if(ys[mail.senderID - 1] == null){
                        ys[mail.senderID - 1] = y;
                        ysCounter ++;
                    }
                }
            }
        }
    }
}
