package JointFeldmanProtocol;

import FeldmanVerifiableSecretSharing.ShamirSecretSharing.Polynomial;

import java.lang.reflect.Type;
import java.math.BigInteger;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;
/**
 * Created by Tzlil on 2/7/2016.
 */
public class Network {

    private final User[] users;
    private final int n;
    private final Queue<Integer> availableIDs;


    public Network(int n) {
        this.n = n;
        this.users = new User[n];
        this.availableIDs = new ArrayBlockingQueue<Integer>(n);
        for (int id = 1; id <= n; id++){
            availableIDs.add(id);
        }
    }

    public User connect(MailHandler messageHandler){
        Integer id = availableIDs.poll();
        if (id == null)
            return null;
        return new User(id,messageHandler);
    }

    private boolean sendMessage(User sender,int destination,Message message){
        if(destination < 1 || destination > n)
            return false;
        User user = users[destination - 1];
        if (user == null)
            return false;
        return user.mailbox.add(new Mail(sender.ID,false,message));
    }

    private void sendBroadcast(User sender,Message message){
        User user;
        int ID = sender.ID;
        for (int i = 0 ; i < n ; i++){
            if (i + 1 == ID) {
                continue;
            }
            user = users[i];
            user.mailbox.add(new Mail(ID,true,message));
        }
    }

    public class User{
        private final MailHandler messageHandler;
        private final Queue<Mail> mailbox;
        private final int ID;

        public User(int ID, MailHandler messageHandler) {
            this.mailbox = new ArrayBlockingQueue<Mail>(n*n);
            this.ID = ID;
            this.messageHandler = messageHandler;
            Thread thread = new Thread(new Receiver());
            thread.run();
        }


        public boolean send(int id, Message message){
            return sendMessage(this,id,message);
        }
        public void broadcast(Message message){
            sendBroadcast(this,message);
        }

        public int getID() {
            return ID;
        }

        private class Receiver implements Runnable{
            @Override
            public void run() {
                while (true){
                    if (!mailbox.isEmpty()){
                        messageHandler.handel(mailbox.poll());
                    }else{
                        try {
                            Thread.sleep(30);
                        } catch (InterruptedException e) {
                            // do nothing
                        }
                    }
                }
            }
        }


    }

    public class Mail{
        public final int senderID;
        public final boolean isPrivate;
        public final Message message;

        private Mail(int senderID, boolean isPrivate, Message message) {
            this.senderID = senderID;
            this.isPrivate = isPrivate;
            this.message = message;
        }
    }

    public static abstract class Message{

    }
    public static class CommitmentMessage extends Message{
        public final int k;
        public final BigInteger commitment;


        public CommitmentMessage(int k, BigInteger commitment) {
            this.k = k;
            this.commitment = commitment;
        }
    }

    public static class YMessage extends Message{

        public final BigInteger y;

        public YMessage(BigInteger y) {
            this.y = y;
        }
    }

    public static class SecretMessage extends Message{
        public final Polynomial.Point secret;

        public SecretMessage(Polynomial.Point secret) {
            this.secret = secret;
        }
    }

    public interface MailHandler {
        public void handel(Mail mail);
    }
}
